@@ -23,7 +23,7 @@
  */
 
 /*
- * mod_ldap_userdir v1.1.16
+ * mod_ldap_userdir v1.1.17
  *
  * Description: A module for the Apache web server that performs UserDir
  * (home directory) lookups from an LDAP directory.
@@ -65,9 +65,9 @@
 #undef strtoul
 #ifdef STANDARD20_MODULE_STUFF
 # define APR_WANT_STRFUNC
-# include "apr_want.h"
-# include "apr_strings.h"
 # include "apr_hash.h"
+# include "apr_strings.h"
+# include "apr_want.h"
 # include <time.h>
 # define AP_POOL apr_pool_t
 # define AP_PSTRDUP apr_pstrdup
@@ -77,6 +77,10 @@
 # define AP_STRSTR_C ap_strstr_c
 # define AP_STRCHR_C ap_strchr_c
 # define AP_TABLE_SETN apr_table_setn
+# define AP_ARRAY_HEADER apr_array_header_t
+# define AP_MAKE_ARRAY apr_array_make
+# define AP_COPY_ARRAY_HDR apr_array_copy_hdr
+# define AP_PUSH_ARRAY apr_array_push
 # if !defined(WIN32) && !defined(OS2) && !defined(BEOS) && !defined(NETWARE)
 #  define HAVE_UNIX_SUEXEC
 #  include "unixd.h"  /* Contains the suexec_identity hook used on Unix */
@@ -90,6 +94,10 @@
 # define AP_STRSTR_C strstr
 # define AP_STRCHR_C strchr
 # define AP_TABLE_SETN ap_table_setn
+# define AP_ARRAY_HEADER array_header
+# define AP_MAKE_ARRAY ap_make_array
+# define AP_COPY_ARRAY_HDR ap_copy_array_hdr
+# define AP_PUSH_ARRAY ap_push_array
 # if !defined(NETWARE)
 #  include <sys/types.h>
 # endif
@@ -102,13 +110,25 @@
 #include <lber.h>
 #include <ldap.h>
 
+#if defined(LDAP_API_FEATURE_X_OPENLDAP) && (LDAP_VENDOR_VERSION >= 192)
+# define HAS_LDAP_UNBIND_EXT_S
+#endif
+
+#if defined(LDAP_API_FEATURE_X_OPENLDAP) && (LDAP_VENDOR_VERSION >= 19905)
+# define HAS_LDAP_INITIALIZE
+#endif
+
 #if LDAP_API_VERSION >= 2000
 # define LDAP_VALUE(values, i) (values[i]->bv_val)
 # define LDAP_VALUE_FREE(values) (ldap_value_free_len(values))
-# define LDAP_UNBIND(ld) (ldap_unbind_ext_s(ld, NULL, NULL))
 #else
 # define LDAP_VALUE(values, i) (values[i])
 # define LDAP_VALUE_FREE(values) (ldap_value_free(values))
+#endif
+
+#ifdef HAS_LDAP_UNBIND_EXT_S
+# define LDAP_UNBIND(ld) (ldap_unbind_ext_s(ld, NULL, NULL))
+#else
 # define LDAP_UNBIND(ld) (ldap_unbind_s(ld, NULL, NULL))
 #endif
 
@@ -125,19 +145,28 @@ module MODULE_VAR_EXPORT ldap_userdir_module;
 
 typedef struct ldap_userdir_config {
 	char *userdir;
+
+	AP_ARRAY_HEADER *userdirs;
+
+	AP_ARRAY_HEADER *servers;
+	unsigned int cur_server_index;
+
 #if LDAP_API_VERSION >= 2000
 	char *url;
 #else /* LDAP_API_VERSION >= 2000 */
 	char *server;
 	int port;
 #endif /* LDAP_API_VERSION >= 2000 */
+
 	char *ldap_dn, *dn_pass,
 	     *basedn, *filter_template,
 	     *home_attr, *username_attr, *uidNumber_attr, *gidNumber_attr;
 	int search_scope, protocol_version;
+
 #ifdef STANDARD20_MODULE_STUFF
 	int cache_timeout;
 #endif
+
 #ifdef TLS
 	int use_tls;
 #endif
@@ -166,23 +195,29 @@ create_ldap_userdir_config(AP_POOL *p, server_rec *s)
 {
 	ldap_userdir_config *newcfg = (ldap_userdir_config *) AP_PCALLOC(p, sizeof(ldap_userdir_config));
 
+	newcfg->userdirs = AP_MAKE_ARRAY(p, 1, sizeof(char *));
+
+	newcfg->servers = AP_MAKE_ARRAY(p, 2, sizeof(char *));
+
 #if LDAP_API_VERSION < 2000
 	newcfg->port = -1;
 #endif /* LDAP_API_VERSION < 2000 */
+
 	newcfg->search_scope = -1;
 	newcfg->protocol_version = -1;
+
 #ifdef STANDARD20_MODULE_STUFF
 	newcfg->cache_timeout = -1;
 #endif
+
 #ifdef TLS
 	newcfg->use_tls = -1;
 #endif /* TLS */
 
-	newcfg->got_url = 0;
-
 #ifdef STANDARD20_MODULE_STUFF
 	newcfg->homedirHt = apr_hash_make(p);
 #endif /* STANDARD20_MODULE_STUFF */
+
 	return (void *)newcfg;
 }
 
@@ -196,9 +231,14 @@ merge_ldap_userdir_config(AP_POOL *p, void *server1_conf, void *server2_conf)
 		(ldap_userdir_config *) AP_PCALLOC(p, sizeof(ldap_userdir_config));
 	memcpy(merged_cfg, s_cfg2, sizeof(ldap_userdir_config));
 
-	if (!merged_cfg->userdir) {
-		merged_cfg->userdir = AP_PSTRDUP(p, s_cfg1->userdir);
+	if (!merged_cfg->userdirs) {
+		merged_cfg->userdirs = AP_COPY_ARRAY_HDR(p, s_cfg1->userdirs);
 	}
+
+	if (!merged_cfg->servers) {
+		merged_cfg->servers = AP_COPY_ARRAY_HDR(p, s_cfg1->servers);
+	}
+
 #if LDAP_API_VERSION >= 2000
 	if (!merged_cfg->url) {
 		merged_cfg->url = AP_PSTRDUP(p, s_cfg1->url);
@@ -207,7 +247,11 @@ merge_ldap_userdir_config(AP_POOL *p, void *server1_conf, void *server2_conf)
 	if (!merged_cfg->server) {
 		merged_cfg->server = AP_PSTRDUP(p, s_cfg1->server);
 	}
+	if (merged_cfg->port == -1) {
+		merged_cfg->port = s_cfg1->port;
+	}
 #endif /* LDAP_API_VERSION >= 2000 */
+
 	if (!merged_cfg->ldap_dn) {
 		merged_cfg->ldap_dn = AP_PSTRDUP(p, s_cfg1->ldap_dn);
 	}
@@ -232,22 +276,20 @@ merge_ldap_userdir_config(AP_POOL *p, void *server1_conf, void *server2_conf)
 	if (!merged_cfg->gidNumber_attr) {
 		merged_cfg->gidNumber_attr = AP_PSTRDUP(p, s_cfg1->gidNumber_attr);
 	}
-#if LDAP_API_VERSION < 2000
-	if (merged_cfg->port == -1) {
-		merged_cfg->port = s_cfg1->port;
-	}
-#endif /* LDAP_API_VERSION < 2000 */
+
 	if (merged_cfg->search_scope == -1) {
 		merged_cfg->search_scope = s_cfg1->search_scope;
 	}
 	if (merged_cfg->protocol_version == -1) {
 		merged_cfg->protocol_version = s_cfg1->protocol_version;
 	}
+
 #ifdef STANDARD20_MODULE_STUFF
 	if (merged_cfg->cache_timeout == -1) {
 		merged_cfg->cache_timeout = s_cfg1->cache_timeout;
 	}
 #endif /* STANDARD20_MODULE_STUFF */
+
 #ifdef TLS
 	if (merged_cfg->use_tls == -1) {
 		merged_cfg->use_tls = s_cfg1->use_tls;
@@ -263,10 +305,10 @@ set_ldap_user_dir(cmd_parms *cmd, void *dummy, const char *arg)
 	ldap_userdir_config *s_cfg = (ldap_userdir_config *) ap_get_module_config(cmd->server->module_config, &ldap_userdir_module);
 
 	if (strlen(arg) == 0) {
-		return "LDAPUserDir must be supplied with the public subdirectory in users' home directories (e.g., 'public_html').";
+		return "LDAPUserDir must be supplied with the public subdirectory in users' home directories (for example, 'public_html' or '.').";
 	}
 
-	s_cfg->userdir = AP_PSTRDUP(cmd->pool, arg);
+	*(char **)AP_PUSH_ARRAY(s_cfg->userdirs) = AP_PSTRDUP(cmd->pool, arg);
 	return NULL;
 }
 
@@ -308,49 +350,32 @@ set_url(cmd_parms *cmd, void *dummy, const char *arg)
 	}
 #endif /* TLS */
 
-#if LDAP_API_VERSION >= 2000
-	s_cfg->url = AP_PSTRDUP(cmd->pool, arg);
-#else
-	if (url->lud_host != NULL) {
-		s_cfg->server = AP_PSTRDUP(cmd->pool, url->lud_host);
-	}
-	if (url->lud_port != 0) {
-		s_cfg->port = url->lud_port;
-	}
-#endif /* LDAP_API_VERSION >= 2000 */
-	if (url->lud_dn != NULL) {
-		s_cfg->basedn = AP_PSTRDUP(cmd->pool, url->lud_dn);
-	}
-	if (url->lud_filter != NULL) {
-		s_cfg->filter_template = AP_PSTRDUP(cmd->pool, url->lud_filter);
-	}
-	if (url->lud_scope != 0) {
-		s_cfg->search_scope = url->lud_scope;
-	}
-
 	ldap_free_urldesc(url);
+
+	*(char **)AP_PUSH_ARRAY(s_cfg->servers) = AP_PSTRDUP(cmd->pool, arg);
 	return NULL;
 }
 
 static const char *
 set_server(cmd_parms *cmd, void *dummy, const char *arg)
 {
+	char *word;
 	ldap_userdir_config *s_cfg = (ldap_userdir_config *) ap_get_module_config(cmd->server->module_config, &ldap_userdir_module);
 
 	if (s_cfg->got_url) {
 		return "LDAPUserDirServer can't be combined with LDAPUserDirServerURL.";
 	}
 
-	if (strlen(arg) == 0) {
-		return "LDAPUserDirServer must be supplied with the name of an LDAP server.";
+	while (*arg) {
+		word = ap_getword_conf(cmd->pool, &arg);
+
+		if (strlen(word) == 0) {
+			return "LDAPUserDirServer must be supplied with the name of an LDAP server.";
+		}
+
+		*(char **)AP_PUSH_ARRAY(s_cfg->servers) = word;
 	}
 
-
-#if LDAP_API_VERSION >= 2000
-	s_cfg->url = AP_PSTRCAT(cmd->pool, "ldap://", arg, "/", NULL);
-#else /* LDAP_API_VERSION >= 2000 */
-	s_cfg->server = AP_PSTRDUP(cmd->pool, arg);
-#endif /* LDAP_API_VERSION >= 2000 */
 	return NULL;
 }
 
@@ -561,7 +586,7 @@ init_ldap_userdir(AP_POOL *pconf, AP_POOL *plog,
 		apply_config_defaults(s_cfg);
 	}
 
-	ap_add_version_component(pconf, "mod_ldap_userdir/1.1.16");
+	ap_add_version_component(pconf, "mod_ldap_userdir/1.1.17");
 	return OK;
 }
 #else /* STANDARD20_MODULE_STUFF */
@@ -574,7 +599,7 @@ init_ldap_userdir(server_rec *s, AP_POOL *p)
 		apply_config_defaults(s_cfg);
 	}
 
-	ap_add_version_component("mod_ldap_userdir/1.1.16");
+	ap_add_version_component("mod_ldap_userdir/1.1.17");
 }
 #endif /* STANDARD20_MODULE_STUFF */
 
@@ -600,14 +625,14 @@ result2errmsg(const request_rec *r, LDAP *ld, LDAPMessage *result)
 }
 
 static int
-connect_ldap_userdir(ldap_userdir_config *s_cfg)
+_ldap_connect(ldap_userdir_config *s_cfg)
 {
 	int ret, sizelimit = 2, version;
 #if LDAP_API_VERSION >= 2000
 	struct berval bindcred;
 #endif
 
-#if LDAP_API_VERSION >= 2000
+#ifdef HAS_LDAP_INITIALIZE
 	ret = ldap_initialize(&(s_cfg->ld), s_cfg->url);
 	if (ret != LDAP_SUCCESS) {
 # ifdef STANDARD20_MODULE_STUFF
@@ -617,7 +642,7 @@ connect_ldap_userdir(ldap_userdir_config *s_cfg)
 # endif
 		return -1;
 	}
-#else /* LDAP_API_VERSION >= 2000 */
+#else /* HAS_LDAP_INITIALIZE */
 	if ((s_cfg->ld = (LDAP *) ldap_init(s_cfg->server, s_cfg->port)) == NULL) {
 # ifdef STANDARD20_MODULE_STUFF
 		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, NULL, "mod_ldap_userdir: ldap_init() to %s failed: %s", s_cfg->server, strerror(errno));
@@ -626,7 +651,7 @@ connect_ldap_userdir(ldap_userdir_config *s_cfg)
 # endif
 		return -1;
 	}
-#endif /* LDAP_API_VERSION >= 2000 */
+#endif /* HAS_LDAP_INITIALIZE */
 
 	switch (s_cfg->protocol_version) {
 	case 2:
@@ -703,6 +728,88 @@ connect_ldap_userdir(ldap_userdir_config *s_cfg)
 	return 1;
 }
 
+static int
+connect_ldap_userdir(request_rec *r, ldap_userdir_config *s_cfg)
+{
+	int start_server_index;
+	char *item;
+	LDAPURLDesc *url;
+
+	start_server_index = s_cfg->cur_server_index;
+	do {
+		/* We won't have any configured servers if no
+		 * LDAPUserDirServer* directives were specified.
+		 * Fall back and use the SDK default if so.
+		 */
+		if (s_cfg->servers->nelts) {
+			item = ((char **)(s_cfg->servers->elts))[s_cfg->cur_server_index];
+
+			if (ldap_is_ldap_url(item)) {
+				if (ldap_url_parse(item, &url) != LDAP_URL_SUCCESS) {
+#ifdef STANDARD20_MODULE_STUFF
+					ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, NULL, "mod_ldap_userdir: URL %s was valid at startup, but is no longer valid?!", item);
+#else /* STANDARD20_MODULE_STUFF */
+					ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, NULL, "mod_ldap_userdir: URL %s was valid at startup, but is no longer valid?!", item);
+#endif /* STANDARD20_MODULE_STUFF */
+
+					++(s_cfg->cur_server_index);
+					if (s_cfg->cur_server_index >= s_cfg->servers->nelts) {
+						s_cfg->cur_server_index = 0;
+					}
+					continue;
+				}
+
+#ifdef HAS_LDAP_INITIALIZE
+				s_cfg->url = AP_PSTRDUP(r->server->process->pool, item);
+#else /* HAS_LDAP_INITIALIZE */
+				if (url->lud_host != NULL) {
+					s_cfg->server = AP_PSTRDUP(r->server->process->pool, url->lud_host);
+				}
+				if (url->lud_port != 0) {
+					s_cfg->port = url->lud_port;
+				}
+#endif /* HAS_LDAP_INITIALIZE */
+				if (url->lud_dn != NULL) {
+					s_cfg->basedn = AP_PSTRDUP(r->server->process->pool, url->lud_dn);
+				}
+				if (url->lud_filter != NULL) {
+					s_cfg->filter_template = AP_PSTRDUP(r->server->process->pool, url->lud_filter);
+				}
+
+				s_cfg->search_scope = url->lud_scope;
+				if (s_cfg->search_scope == LDAP_SCOPE_BASE) {
+# ifdef STANDARD20_MODULE_STUFF
+					ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, 0, NULL, "mod_ldap_userdir: WARNING: LDAP URL search scopes default to 'base' (not 'sub') and may not be what you want.");
+# else
+					ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, NULL, "mod_ldap_userdir: WARNING: LDAP URL search scopes default to 'base' (not 'sub') and may not be what you want.");
+# endif
+				}
+
+				ldap_free_urldesc(url);
+			} else {
+#ifdef HAS_LDAP_INITIALIZE
+				s_cfg->url = AP_PSTRCAT(r->server->process->pool,
+					"ldap://", item, "/", NULL);
+#else /* HAS_LDAP_INITIALIZE */
+				s_cfg->server = AP_PSTRDUP(r->server->process->pool, item);
+				s_cfg->port = LDAP_PORT;
+#endif /* HAS_LDAP_INITIALIZE */
+			}
+		}
+
+		if (_ldap_connect(s_cfg) == 1) {
+			return 1;
+		}
+
+		++s_cfg->cur_server_index;
+		if (s_cfg->cur_server_index >= s_cfg->servers->nelts) {
+			s_cfg->cur_server_index = 0;
+		}
+	} while (s_cfg->cur_server_index != start_server_index);
+
+	return -1;
+}
+
 static char *
 generate_filter(AP_POOL *p, char *template, const char *entity)
 {
@@ -818,7 +925,7 @@ get_ldap_homedir(ldap_userdir_config *s_cfg, request_rec *r,
 	/* If we haven't even connected yet, try to connect. If we still can't
 	   connect, give up. */
 	if (s_cfg->ld == NULL) {
-		if (connect_ldap_userdir(s_cfg) != 1) {
+		if (connect_ldap_userdir(r, s_cfg) != 1) {
 			return NULL;
 		}
 	}
@@ -845,7 +952,7 @@ get_ldap_homedir(ldap_userdir_config *s_cfg, request_rec *r,
 		if (ret == LDAP_SERVER_DOWN) {
 			LDAP_UNBIND(s_cfg->ld);
 
-			if (connect_ldap_userdir(s_cfg) != 1) {
+			if (connect_ldap_userdir(r, s_cfg) != 1) {
 #ifdef STANDARD20_MODULE_STUFF
 				ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, 0, r, "mod_ldap_userdir: LDAP server went away, couldn't reconnect. Declining request.");
 #else
@@ -999,16 +1106,15 @@ translate_ldap_userdir(request_rec *r)
 {
 	const char *w, *dname;
 	char *name = r->uri;
+	int userdir_index;
 	const ldap_userdir_config *s_cfg = (ldap_userdir_config *) ap_get_module_config(r->server->module_config, &ldap_userdir_module);
-	const char *userdirs = s_cfg->userdir;
 	request_rec *notes_req;
 	struct hash_entry *user_info;
 
-	/*
-	 * If the URI doesn't match our basic pattern, we've nothing to do with
-	 * it.
+	/* If the URI doesn't match our basic pattern, we've nothing to do
+	 * with it.
 	 */
-	if ((s_cfg->userdir == NULL) ||
+	if ((s_cfg->userdirs->nelts == 0) ||
 	    (name[0] != '/') ||
 	    (name[1] != '~'))
 	{
@@ -1018,8 +1124,7 @@ translate_ldap_userdir(request_rec *r)
 	dname = name + 2;
 	w = ap_getword(r->pool, &dname, '/');
 
-	/*
-	 * This 'dname' funny business involves backing it up to capture the '/'
+	/* This 'dname' funny business involves backing it up to capture the '/'
 	 * delimiting the "/~user" part from the rest of the URL, in case there
 	 * was one (the case where there wasn't being just "GET /~user HTTP/1.0",
 	 * for which we don't want to tack on a '/' onto the filename).
@@ -1028,9 +1133,7 @@ translate_ldap_userdir(request_rec *r)
 		--dname;
 	}
 
-	/*
-	 * If there's no username, it's not for us. Ignore . and .. as well.
-	 */
+	/* If there's no username, it's not for us. Ignore . and .. as well. */
 	if (w[0] == '\0' || (w[1] == '.' && (w[2] == '\0' || (w[2] == '.' && w[3] == '\0')))) {
 		return DECLINED;
 	}
@@ -1040,8 +1143,10 @@ translate_ldap_userdir(request_rec *r)
 		return DECLINED;
 	}
 
-	while (*userdirs) {
-		const char *userdir = ap_getword_conf(r->pool, &userdirs);
+	userdir_index = 0;
+	while (userdir_index < s_cfg->userdirs->nelts) {
+		const char *userdir = ((char **)(s_cfg->userdirs->elts))[userdir_index];
+		++userdir_index;
 		char *filename;
 #ifdef STANDARD20_MODULE_STUFF
 		apr_finfo_t statbuf;
@@ -1085,7 +1190,7 @@ translate_ldap_userdir(request_rec *r)
 		 */
 #ifdef STANDARD20_MODULE_STUFF
 		if (filename &&
-		    (!*userdirs ||
+		    (userdir_index >= s_cfg->userdirs->nelts ||
 		     ((rv = apr_stat(&statbuf, filename, APR_FINFO_MIN, r->pool)) == APR_SUCCESS ||
 		      rv == APR_INCOMPLETE)))
 #else
@@ -1096,7 +1201,7 @@ translate_ldap_userdir(request_rec *r)
 			/* When statbuf contains info on r->filename, we can save
 			 * a syscall by copying it to r->finfo.
 			 */
-			if (*userdirs && dname[0] == 0) {
+			if (userdir_index < s_cfg->userdirs->nelts && dname[0] == 0) {
 				r->finfo = statbuf;
 			}
 
@@ -1200,11 +1305,11 @@ register_hooks(AP_POOL *p)
 
 static const command_rec ldap_userdir_cmds[] = {
 #ifdef STANDARD20_MODULE_STUFF
-	AP_INIT_RAW_ARGS("LDAPUserDir", set_ldap_user_dir, NULL, RSRC_CONF,
+	AP_INIT_ITERATE("LDAPUserDir", set_ldap_user_dir, NULL, RSRC_CONF,
 	                 "the public subdirectory in users' home directories"),
-	AP_INIT_TAKE1("LDAPUserDirServer", set_server, NULL, RSRC_CONF,
+	AP_INIT_ITERATE("LDAPUserDirServer", set_server, NULL, RSRC_CONF,
 	              "the LDAP directory server that will be used for LDAP UserDir queries"),
-	AP_INIT_TAKE1("LDAPUserDirServerURL", set_url, NULL, RSRC_CONF,
+	AP_INIT_ITERATE("LDAPUserDirServerURL", set_url, NULL, RSRC_CONF,
 	              "the LDAP URL that will be used for LDAP UserDir queries"),
 	AP_INIT_TAKE2("LDAPUserDirDNInfo", set_ldap_dninfo, NULL, RSRC_CONF,
 	              "the DN and password that will be used to bind to the LDAP server when doing LDAP UserDir lookups"),
@@ -1223,11 +1328,11 @@ static const command_rec ldap_userdir_cmds[] = {
 	AP_INIT_TAKE1("LDAPProtocolVersion", set_ldap_protocol_version, NULL, RSRC_CONF,
 	              "the LDAP protocol version to use"),
 #else /* STANDARD20_MODULE_STUFF */
-	{"LDAPUserDir", set_ldap_user_dir, NULL, RSRC_CONF, RAW_ARGS,
+	{"LDAPUserDir", set_ldap_user_dir, NULL, RSRC_CONF, ITERATE,
 	 "the public subdirectory in users' home directories"},
-	{"LDAPUserDirServer", set_server, NULL, RSRC_CONF, TAKE1,
+	{"LDAPUserDirServer", set_server, NULL, RSRC_CONF, ITERATE,
 	 "the LDAP directory server that will be used for LDAP UserDir queries"},
-	{"LDAPUserDirServerURL", set_url, NULL, RSRC_CONF, TAKE1,
+	{"LDAPUserDirServerURL", set_url, NULL, RSRC_CONF, ITERATE,
 	 "the LDAP URL that will be used for LDAP UserDir queries"},
 	{"LDAPUserDirDNInfo", set_ldap_dninfo, NULL, RSRC_CONF, TAKE2,
 	 "the DN and password that will be used to bind to the LDAP server when doing LDAP UserDir lookups"},
